
#ifndef PLS_DATAFLOW_INTERNAL_FUNCTION_NODE_H_
#define PLS_DATAFLOW_INTERNAL_FUNCTION_NODE_H_

#include <tuple>
#include <iostream>
#include <atomic>

#include "in_port.h"
#include "out_port.h"
#include "node.h"

#include "inputs.h"
#include "outputs.h"

#include "pls/internal/helpers/seqence.h"
#include "pls/pls.h"

namespace pls {
namespace dataflow {
namespace internal {

using namespace pls::internal::helpers;

template<typename INS, typename OUTS, typename F>
class function_node {};

template<typename I0, typename ...I, typename O0, typename ...O, typename F>
class function_node<pls::dataflow::inputs<I0, I...>, pls::dataflow::outputs<O0, O...>, F> : public node {
 private:
  // Our own type
  using self_type = function_node<pls::dataflow::inputs<I0, I...>, pls::dataflow::outputs<O0, O...>, F>;

  // Input-Output port types
  using multi_in_port_type = multi_in_port<self_type, 0, I0, I...>;
  using multi_out_port_type = multi_out_port<O0, O...>;

  // Input-Output value tuples
  using input_tuple = std::tuple<token<I0>, token<I>...>;
  using output_tuple = std::tuple<token<O0>, token<O>...>;
  static constexpr int num_in_ports = std::tuple_size<input_tuple>();
  static constexpr int num_out_ports = std::tuple_size<output_tuple>();

  // Memory used for ONE active invocation of the dataflow-graph
  struct invocation_memory {
    std::atomic<int> inputs_missing_;
    input_tuple input_buffer_;
  };

 public:
  template<int POS>
  using in_port_at = typename multi_in_port_type::template in_port_type_at<POS>;
  template<int POS>
  using out_port_at = typename multi_out_port_type::template out_port_type_at<POS>;

  template<int POS>
  in_port_at<POS> &in_port() {
    return in_port_.template get<POS>();
  }

  template<int POS>
  out_port_at<POS> &out_port() {
    return out_port_.template get<POS>();
  }

  template<int POS, typename T>
  void token_pushed(token<T> token) {
    auto current_memory = get_invocation<invocation_memory>(token.invocation());

    std::get<POS>(current_memory->input_buffer_) = token;
    auto remaining_inputs = --(current_memory->inputs_missing_);
    if (remaining_inputs == 0) {
      execute_function(current_memory, token.invocation());
      current_memory->inputs_missing_ = num_in_ports;
    }
  }

  void execute_function(invocation_memory *invocation_memory, invocation_info invocation_info) {
    auto lambda = [&]() {
      input_tuple &inputs = invocation_memory->input_buffer_;
      output_tuple outputs;

      execute_function_internal(inputs, typename sequence_gen<1 + sizeof...(I)>::type(),
                                outputs, typename sequence_gen<1 + sizeof...(O)>::type(), invocation_info);
    };
    // TODO: maybe replace this with 'continuation' style invocation
    pls::scheduler::spawn_child_and_wait<pls::lambda_task_by_reference<decltype(lambda)>>(lambda);
  }
  template<typename T>
  void set_invocation_info(token<T> &token, invocation_info invocation_info) {
    token.set_invocation(invocation_info);
  }
  template<int N, typename ...OT>
  struct propagate_output {
    propagate_output(multi_out_port_type &, output_tuple &) {}
    void propagate() {}
  };
  template<int N, typename OT1, typename ...OT>
  struct propagate_output<N, OT1, OT...> {
    multi_out_port_type &out_port_;
    output_tuple &output_tuple_;

    propagate_output(multi_out_port_type &out_port, output_tuple &output_tuple) : out_port_{out_port},
                                                                                  output_tuple_{output_tuple} {}

    void propagate() {
      out_port_.template get<N>().push_token(std::get<N>(output_tuple_));
      propagate_output<N + 1, OT...>{out_port_, output_tuple_}.propagate();
    }
  };

  template<int ...IS, int ...OS>
  void execute_function_internal(input_tuple &inputs,
                                 sequence<IS...>,
                                 output_tuple &outputs,
                                 sequence<OS...>,
                                 invocation_info invocation_info) {
    set_invocation_info(std::get<OS>(outputs)..., invocation_info);
    function_(std::get<IS>(inputs).value()..., std::get<OS>(outputs).value()...);
    propagate_output<0, O0, O...>{out_port_, outputs}.propagate();
  }

  explicit function_node(F function) : in_port_{this, this}, out_port_{}, function_{function} {}

  //////////////////////////////////////////////////////////////////
  // Overrides for generic node functionality (building the graph)
  //////////////////////////////////////////////////////////////////
  int num_successors() const override {
    return num_out_ports;
  }
  node *successor_at(int pos) const override {
    return out_port_.next_node_at(pos);
  }

  bool is_fully_connected() const override {
    return in_port_.fully_connected() && out_port_.fully_connected();
  }

  int instance_buffer_size() const override {
    return sizeof(invocation_memory);
  }
  void init_instance_buffer(void *memory) const override {
    auto invocation = new(memory) invocation_memory{};
    invocation->inputs_missing_ = num_in_ports;
  };

 private:
  multi_in_port_type in_port_;
  multi_out_port_type out_port_;

  F function_;
};

}
}
}

#endif //PLS_DATAFLOW_INTERNAL_NODE_H_
